{********************************************************************}
{* Common functions and procedures unit for LogicGates5 application *}
{********************************************************************}

unit common;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, StdCtrls, ExtCtrls, Graphics, Dialogs;

type
  // Switches description
  TSwitch = record
    SwitchShapeA, SwitchShapeB: TShape;                                        // horizontal and vertical parts (shapes) of the switch
    SwitchShapeAOffPos: Integer;                                               // vertical position ("Top") of the horiz. switch part, when switch is off
    SwitchOn: Boolean;                                                         // switch on (True) or off (False)
  end;
  // LEDs description
  TLed = record
    LedShape: TShape;                                                          // shape of the LED
    LedOn: Boolean;                                                            // LED on (True) or off (False)
  end;

// Public procedures for SR and D-type latches simulation (called from the different simulation windows)

procedure LedOn(var Led: TLed);
procedure LedOff(var Led: TLed);
procedure SwitchOn(var Switch: TSwitch);
procedure SwitchOff(var Switch: TSwitch);
procedure Latch_SR_nand(SwitchAction: string; Sbar, Rbar: Boolean; var Q, Qbar: Boolean; var LedQ, LedQbar: TLed; out Forbidden: Boolean);
procedure Latch_SR_nand2(SwitchAction: string; S, R: Boolean; var Q, Qbar: Boolean; var LedQ, LedQbar: TLed; out Forbidden: Boolean);
procedure Latch_RS_nor(SwitchAction: string; R, S: Boolean; var Q, Qbar: Boolean; var LedQ, LedQbar: TLed; out Forbidden: Boolean);
procedure Latch_SR_gated_nand(SwitchAction: string; E, S, R: Boolean; var Q, Qbar: Boolean; var LedLSbar, LedLRbar, LedQ, LedQbar: TLed; out Forbidden: Boolean);
procedure Latch_RS_gated_nor(SwitchAction: string; E, R, S: Boolean; var Q, Qbar: Boolean; var LedLR, LedLS, LedQ, LedQbar: TLed; out Forbidden: Boolean);
procedure Latch_D_nand(SwitchAction: string; E, D: Boolean; var Q, Qbar: Boolean; var LedLSbar, LedLRbar, LedQ, LedQbar: TLed);
procedure Latch_D_nand2(SwitchAction: string; E, D: Boolean; var Q, Qbar: Boolean; var LedLSbar, LedLRbar, LedQ, LedQbar: TLed);
procedure Latch_D_nor(SwitchAction: string; E, D: Boolean; var Q, Qbar: Boolean; var LedLR, LedLS, LedQ, LedQbar: TLed);

implementation

// Switches related constants
const
  SwitchShapeBOffPos = -15;                                                    // offset of vertical position ("Top") of the vert. switch part
  SwitchShapeOnPos = +12;                                                      // offset of vertical position ("Top") of the switch parts, when switch is on

{ Turn a given switch on }

procedure SwitchOn(var Switch: TSwitch);

begin
  Switch.SwitchOn := True;
  Switch.SwitchShapeA.Top := Switch.SwitchShapeAOffPos + SwitchShapeOnPos;     // vertical position of horizontal switch part
  Switch.SwitchShapeB.Top := Switch.SwitchShapeA.Top + SwitchShapeBOffPos;     // vertical position of vertical switch part
end;

{ Turn a given switch off }

procedure SwitchOff(var Switch: TSwitch);

begin
  Switch.SwitchOn := False;
  Switch.SwitchShapeA.Top := Switch.SwitchShapeAOffPos;                        // vertical position of horizontal switch part
  Switch.SwitchShapeB.Top := Switch.SwitchShapeA.Top + SwitchShapeBOffPos;     // vertical position of vertical switch part
end;

{ Turn a given LED on }

procedure LedOn(var Led: TLed);

begin
  Led.LedOn := True;
  Led.LedShape.Brush.Style := bsSolid;                                         // LED on is shown by solid brush style (fully colored shaped)
end;

{ Turn a given LED off }

procedure LedOff(var Led: TLed);

begin
  Led.LedOn := False;
  Led.LedShape.Brush.Style := bsDiagCross;                                     // LED on is shown by diagonal cross brush style (sparsely colored shaped)
end;

{ Turn LED on or off, depending on given logic value }

procedure LedOnOff(TurnLedOn: Boolean; var Led: TLed);

begin
  if TurnLedOn then
    LedOn(Led)
  else
    LedOff(Led);
end;

{ Active loww NAND gates SR latch simulation }

procedure Latch_SR_nand(SwitchAction: string; Sbar, Rbar: Boolean; var Q, Qbar: Boolean; var LedQ, LedQbar: TLed; out Forbidden: Boolean);

begin
  Forbidden := False;
  if (not Sbar) and (not Rbar) then
    // Inputs should not be both logic 0
    Forbidden := True;
  // Set outputs Q and Qbar by performing a NAND with Sbar and Qbar resp. Rbar and Q
  // Because of the output feedback on the inputs of both NAND gates, the program must first calculate the output
  // corresponding to the input, that has been set. This is done by considering a string argument, that tells the
  // routine, which switch has been pushed by the user.
  if SwitchAction = 'set' then begin
    Q := not (Sbar and Qbar);                                                  // Q    = Sbar NAND Qbar
    Qbar := not (Rbar and Q);                                                  // Qbar = Rbar NAND Q
  end
  else begin
    Qbar := not (Rbar and Q);
    Q := not (Sbar and Qbar);
  end;
  // Turn output LEDs on/off, depending on logic value at the output
  LedOnOff(Q, LedQ);
  LedOnOff(Qbar, LedQbar);
end;

{ Active high NAND gates SR latch simulation }

procedure Latch_SR_nand2(SwitchAction: string; S, R: Boolean; var Q, Qbar: Boolean; var LedQ, LedQbar: TLed; out Forbidden: Boolean);

begin
  Forbidden := False;
  if S and R then
    // Inputs should not be both logic 1
    Forbidden := True;
  // Active high NAND gates SR latch = active low NAND gates SR latch with both inputs inverted
  // This is true, if you have a look at the circuit and also for the program code...
  Latch_SR_nand(SwitchAction, not S, not R, Q, Qbar, LedQ, LedQbar, Forbidden);
end;

{ Active high NOR gates RS latch simulation }

procedure Latch_RS_nor(SwitchAction: string; R, S: Boolean; var Q, Qbar: Boolean; var LedQ, LedQbar: TLed; out Forbidden: Boolean);

begin
  Forbidden := False;
  if R and S then
    // Inputs should not be both logic 1
    Forbidden := True;
  // Set outputs Q and Qbar by performing a Nor with R and Qbar resp. S and Q (Q corresponding to R -> cf. circuit)
  // Because of the output feedback on the inputs of both NOR gates, the program must first calculate the output
  // corresponding to the input, that has been set. This is done by considering a string argument, that tells the
  // routine, which switch has been pushed by the user.
  if SwitchAction = 'reset' then begin
    Q := not (R or Qbar);                                                      // Q    = R NOR Qbar
    Qbar := not (S or Q);                                                      // Qbar = S NOR Q
  end
  else begin
    Qbar := not (S or Q);
    Q := not (R or Qbar);
  end;
  // Turn output LEDs on/off, depending on logic value at the output
  LedOnOff(Q, LedQ);
  LedOnOff(Qbar, LedQbar);
end;

{ Gated NAND gates SR latch simulation }

procedure Latch_SR_gated_nand(SwitchAction: string; E, S, R: Boolean; var Q, Qbar: Boolean; var LedLSbar, LedLRbar, LedQ, LedQbar: TLed; out Forbidden: Boolean);

var
  Latch_Sbar, Latch_Rbar: Boolean;

begin
  // Determine the logic value of the SR latch inputs Sbar and Rbar
  // They are the result of a NAND of S and E resp. R and E (S and R being the logic values set by the SET / RESET switch)
  Latch_Sbar := not (S and E); Latch_Rbar := not (R and E);
  // Turn the corresponding LEDs on or off
  if Latch_Sbar then
    LedOn(LedLSbar)
  else
    LedOff(LedLSbar);
  if Latch_Rbar then
    LedOn(LedLRbar)
  else
    LedOff(LedLRbar);
  // The issue, caused by the "double feedback" is resolved in the basic SR latch routine, if the SET or RESET switch has been activated,
  // but not, if it has been the ENABLE switch. The action in this case may be either SET or RESET (with one or the other output hving to
  // be calculated first). Considering the basic SR gate's inputs: If Rbar is low, the action to be performed is a RESET...
  if SwitchAction = 'enable' then begin
    if not Latch_Rbar then
      SwitchAction := 'reset'
    else
      SwitchAction := 'set';
  end;
  // Basic active low NAND gates SR latch (with logic input values as calculated)
  Latch_SR_nand(SwitchAction, Latch_Sbar, Latch_Rbar, Q, Qbar, LedQ, LedQbar, Forbidden);
end;

{ Gated NOR gates RS latch simulation }

procedure Latch_RS_gated_nor(SwitchAction: string; E, R, S: Boolean; var Q, Qbar: Boolean; var LedLR, LedLS, LedQ, LedQbar: TLed; out Forbidden: Boolean);

var
  Latch_S, Latch_R: Boolean;

begin
  // Determine the logic value of the RS latch inputs R and S
  // They are the result of a AND of R and E resp. S and E (S and R being the logic values set by the SET / RESET switch)
  Latch_R := R and E; Latch_S := S and E;
  // Turn the corresponding LEDs on or off
  if Latch_R then
    LedOn(LedLR)
  else
    LedOff(LedLR);
  if Latch_S then
    LedOn(LedLS)
  else
    LedOff(LedLS);
  // Determine if action will be SET or RESET (cf. gated NAND gates SR latch for details)
  if SwitchAction = 'enable' then begin
    if Latch_S then
      SwitchAction := 'set'
    else
      SwitchAction := 'reset';
  end;
  // Basic active high NOR gates RS latch (with logic input values as calculated)
  Latch_RS_nor(SwitchAction, Latch_R, Latch_S, Q, Qbar, LedQ, LedQbar, Forbidden);
end;

{ NAND gates D-type latch simulation }

procedure Latch_D_nand(SwitchAction: string; E, D: Boolean; var Q, Qbar: Boolean; var LedLSbar, LedLRbar, LedQ, LedQbar: TLed);

var
  Latch_Sbar, Latch_Rbar: Boolean;
  Forbidden: Boolean;                                                // not used for D-type latches

begin
  // Determine the logic value of the RS latch inputs Sbar and Rbar
  // They are the result of a NAND of D and E resp. Dbar (not D) and E
  Latch_Sbar := not (D and E); Latch_Rbar := not((not D) and E);
  // Turn the corresponding LEDs on or off
  if Latch_Sbar then
    LedOn(LedLSbar)
  else
    LedOff(LedLSbar);
  if Latch_Rbar then
    LedOn(LedLRbar)
  else
    LedOff(LedLRbar);
  // Determine if action will be SET or RESET (cf. gated NAND gates SR latch for details)
  if SwitchAction = 'enable' then begin
    if not Latch_Rbar then
      SwitchAction := 'reset'
    else
      SwitchAction := 'set';
  end;
  // Basic active low NAND gates SR latch (with logic input values as calculated)
  Latch_SR_nand(SwitchAction, Latch_Sbar, Latch_Rbar, Q, Qbar, LedQ, LedQbar, Forbidden);
end;

{ NAND gates D-type latch (variant) simulation }

procedure Latch_D_nand2(SwitchAction: string; E, D: Boolean; var Q, Qbar: Boolean; var LedLSbar, LedLRbar, LedQ, LedQbar: TLed);

var
  Latch_Sbar, Latch_Rbar: Boolean;
  Forbidden: Boolean;                                                          // not used for D-type latches

begin
  // Determine the logic value of the RS latch inputs Sbar and Rbar
  // They are the result of a NAND of D and E resp. the output of this (what equals Dbar) and E
  Latch_Sbar := not (D and E); Latch_Rbar := not(Latch_Sbar and E);
  // Turn the corresponding LEDs on or off
  if Latch_Sbar then
    LedOn(LedLSbar)
  else
    LedOff(LedLSbar);
  if Latch_Rbar then
    LedOn(LedLRbar)
  else
    LedOff(LedLRbar);
  // Determine if action will be SET or RESET (cf. gated NAND gates SR latch for details)
  if SwitchAction = 'enable' then begin
    if not LedLSbar.LedOn then
      SwitchAction := 'reset'
    else
      SwitchAction := 'set';
  end;
  // Basic active low NAND gates SR latch (with logic input values as calculated)
  Latch_SR_nand(SwitchAction, Latch_Sbar, Latch_Rbar, Q, Qbar, LedQ, LedQbar, Forbidden);
end;

{ NOR gates D-type latch simulation }

procedure Latch_D_nor(SwitchAction: string; E, D: Boolean; var Q, Qbar: Boolean; var LedLR, LedLS, LedQ, LedQbar: TLed);

var
  Latch_S, Latch_R: Boolean;
  Forbidden: Boolean;                                                          // not used for D-type latches

begin
  // Determine the logic value of the RS latch inputs S and R
  // They are the result of a NAND of D and E resp. the output of this (what equals Dbar) and E
  Latch_R := (not D) and E; Latch_S := D and E;
  // Turn the corresponding LEDs on/off
  if Latch_R then
    LedOn(LedLR)
  else
    LedOff(LedLR);
  if Latch_S then
    LedOn(LedLS)
  else
    LedOff(LedLS);
  // Determine if action will be SET or RESET (cf. gated NAND gates SR latch for details)
  if SwitchAction = 'enable' then begin
    if Latch_S then
      SwitchAction := 'set'
    else
      SwitchAction := 'reset';
  end;
  // Basic active high NOR gates SR latch (with logic input values as calculated)
  Latch_RS_nor(SwitchAction, Latch_R, Latch_S, Q, Qbar, LedQ, LedQbar, Forbidden);
end;

end.

